# version of pmzbs to be launched under awk

# Petr Mikulik's Zip Backup System 

# The date of this version: see below
# location on the Internet: http://www.sci.muni.cz/~mikulik/


BEGIN {
pmzbsVersion="29.6.1999"

# This init is required:
OS2=0	# operating system: 1 for IBM OS/2, otherwise Unix
	# this is needed for system-depend. command using slashes or backslashes, 
	# and the names of some command

if (OS2) { 
	GNUpatch="gpatch"	# name of the GNU patch program
	mkdir_with_parents="e:/unix/mkdir.exe -p"
				# there must be the exact path to the GNU mkdir,
				# otherwise the OS/2 mkdir would be used
	# mkdir_with_parents="" # comment this out if you do not have the above
	UNZIP="unzip -C"	# unzip with case-insensitive option
	NULL="nul"		# null device
	ZIP_l_FILE=6		# which columns is the File in `unzip -l` format
	BACKSLASHES=1		# directory names use backslashes
	}
else	{
	GNUpatch="patch"	# name of the GNU patch program
	mkdir_with_parents="mkdir -p"
				# GNU mkdir it allows the parameter -p
	# mkdir_with_parents="" # comment this out if you do not have the above
	UNZIP="unzip"		# unzip (no option, case-sensitive filesystem)
	NULL="/dev/null"	# null device
	ZIP_l_FILE=4		# which columns is the File in `unzip -l` format
	BACKSLASHES=0		# directory names use forward slashes
	}

Today=strftime("%d.%m.%Y",systime())
TodayTime=Today" "strftime("%H:%M:%S",systime())

print "\nThis is Petr Mikulik's Zip Backup System, version "pmzbsVersion"\n"
if (ARGC==1) { Help(); exit }

Repository=ENVIRON["PMZBSROOT"]
gsub("\\\\","/",Repository) # convert into the Unix format (forward slashes)
if (Repository=="") 
  ERROR("environmental variable PMZBSROOT not established")
if (!DirectoryExist(Repository))
  ERROR("\tRepository directory "Repository" does not exist.\n\tPlease create it manually.\n")


# pmzbs list [_project(s)_]
if (ARGV[1]=="list") {
  list="ls -1 -F -R "Repository" 2>"NULL" | grep \":$\""
  print "Projects found in the repository "Repository":"
  do {
    flag = (list | getline S)
    S=substr(S,1,length(S)-1)
    if (FileExist(S"/.Versions")) {
	S=substr(S,length(Repository)+2) # take away the full path
	if (ARGC==2) ProcFor_list(S)  # no parameters given
	  else	{ #
		for (i=2; i<ARGC; i++)
		  if (match(S,ARGV[i])!=0) { 
		    ProcFor_list(S)
		    break }
		}
	}
    }
  while (flag>0)
  close(list)
  exit
  } # end of  pmzbs list [_project(s)_]

if (ARGV[1]=="backup" || ARGV[1]=="backup_newbranch") {
    S=".pmzbs" # Project name is the in the file ;pmzbs
    flag = (getline Project < S);  close(S)
    if (flag<=0) ERROR("Cannot read file .pmzbs !")
    }
  else # Project name is the ARGV[2]
    Project=ARGV[2]
gsub("\\\\","/",Project) # convert into the Unix format (forward slashes)
if (substr(Project,1,1)=="/" )
  ERROR("Project name cannot start with slash!")
ProjectDir=Repository"/"Project
#print "@@PROJECT="ProjectDir"@@"

# these files starting by "," can be changed by the user
FilesFile=ProjectDir"/,Files"
CurrVersionFile=ProjectDir"/,CurrVersion"
# these files starting by ";" must not be changed by the user
LastVersionFile=ProjectDir"/.LastVersion"
VersionsFile=ProjectDir"/.Versions"
StartLevelsFile=ProjectDir"/.StartLevels"

# pmzbs new _project_
if (ARGV[1]=="new") {
  if (ARGC!=4) ERROR("pmzbs new  requires two arguments")
  VersionLevels=ARGV[3]
  if (VersionLevels<1) ERROR("version levels must be positive")
  exist=DirectoryExist(ProjectDir)
  if (exist) # Projects's repository exists, thus:
	ERROR("The repository for the project: "Project" already exists => NOT changed.")
  print "...creating the repository for the project: "Project
  if (index(Project,"/")==0) # now slashes (structure) in the project
	MakeDir(ProjectDir) # create its directory
    else {
	if (mkdir_with_parents!="")  # create its directory
		system(mkdir_with_parents" "ProjectDir)
	else	MakeDir(ProjectDir)
	if (!DirectoryExist(ProjectDir))
	  ERROR("\nCANNOT CREATE DIRECTORY STRUCTURE: either use GNU mkdir, or create the parent\ndirectories yourself.")
	}
  print "# Sequence of the versions of the project: "Project"\n# Created "TodayTime"\n" >VersionsFile
  close(VersionsFile)
  # now create the ",pmzbs" file in the current directory
  S=".pmzbs"
  print Project >S
  print "\nPlease do not modify the first line of this file." >S
  close (S)
  if (OS2) system("attrib +R +H .pmzbs")
    else system("chmod -w .pmzbs")
  # now create the file storing the starting number of levels
  print VersionLevels >StartLevelsFile
  close (StartLevelsFile)
  # now create the file ",CurrVersion" storing the current version
  CurrVersion=""
  for (VersionLevels--; VersionLevels>0; VersionLevels--) CurrVersion=CurrVersion"0."
  CurrVersion=CurrVersion"01"
  MakeCurrVersionFile(CurrVersion)
  # print "# Files to be backuped for the current project: "Project"\n# Created "TodayTime >FilesFile
  print "# Syntax of this file: each file on a separate line, or any INFO-Zip commands\n" >FilesFile
  close(FilesFile)
  print "...done."
  if (BACKSLASHES) gsub("/","\\",FilesFile) # convert into the backslash format
  print "\nPlease, edit the file "FilesFile"\n  and write there all the files (or INFO-Zip commands) for backuping"
  exit
  }
  # end of: pmzbs new _project_


# pmzbs backup _project_   and   pmzbs backup _project_
if (ARGV[1]=="backup" || ARGV[1]=="backup_newbranch") {
  # if (ARGC!=3) ERROR("pmzbs backup  requires an argument")
  if (ARGC!=2) ERROR("pmzbs backup  does not take any argument")
  ProjectMustExist(Project,ProjectDir)
  print "Backuping project: "Project
  # read the current version CurrVersion as the first line at the file CurrVersionFile
  tmp=(getline CurrVersion <CurrVersionFile)
  if (tmp<0) FATAL_ERROR("Missing current version file "CurrVersionFile)
  close(CurrVersionFile)
  if (CurrVersion=="") FATAL_ERROR("Current version file "CurrVersionFile" does not contain any version")
  # now verify that this version does not already exist
  if (VersionExists(CurrVersion)) 
     ERROR(" version "CurrVersion" specified in the file\n\t"CurrVersionFile"\nalready exists (see the file "VersionsFile" and "Version".zip).\nPlease, choose another version name.")
  # now backup to this version

  # what is the latest backuped version?
  if (ARGV[1]=="backup_newbranch")
	BackupAll=1
    else {
	flag=(getline LastVersion<LastVersionFile)
	close(LastVersionFile)
	BackupAll = (flag!=1)
	}
  if (BackupAll) print "...backuping all files from the project "
	    else print "...the last backuped version is:\t"LastVersion
  print "...the current version is:\t\t"CurrVersion
  # well, now do the backup
  DirLastVersion=ProjectDir"/"LastVersion
  DirCurrVersion=ProjectDir"/"CurrVersion
  LastZipFile=ProjectDir"/"LastVersion".zip"
  CurrZipFile=ProjectDir"/"CurrVersion".zip"
  ## Older: each zip in its directory
  ## LastZipFile=DirLastVersion"/"CurrVersion".zip"
  ## CurrZipFile=DirCurrVersion"/"CurrVersion".zip"
  ## print "Creating directory "DirCurrVersion
  ## MakeDir(DirCurrVersion)


  # now zip (=backup) everything in the ",Files" file
  print "...zipping the current version ->\t"CurrVersion".zip"
  do { # while something in the file
    flag=getline0_nocomment(FilesFile)
    if (flag>0 && length($0)>0) { # zip it
	system("zip -o -D "CurrZipFile" "$0" 2>"NULL)
	}
    }
  while (flag>0)
  close(FilesFile)

  # was it ok?
  if (!FileExist(CurrZipFile)) ERROR(CurrZipFile" not created.\nPlease check and edit the file "FilesFile)

  if (!BackupAll) {
	# now change the backup of the previous version wrt the current backup
	# unzip the Last zip into the directory DirLastVersion
	print "...diff-ing the backup of the previous version ("LastVersion".zip)"
	MakeDir(DirLastVersion)
	system(UNZIP" -q -o "LastZipFile" -d "DirLastVersion)

	# now go over each file in the LastZipFile and make diff with the current version
	ListOfLastFiles=UNZIP" -l "LastZipFile
	ListOfLastFiles | getline; ListOfLastFiles | getline # skip 3 lines
	ListOfLastFiles | getline; ListOfLastFiles | getline
	while (substr($1,1,3)!="---") {
	  File=$ZIP_l_FILE
	  if (!FileExist(File)) 
		print "...this file is no more present: "File
	    else { # make diff only if the File is still in the current release
		print "  diffing "File
		system("diff "File" "DirLastVersion"/"File" >\""DirLastVersion"/"File";diff\"")
		system("rm -f "DirLastVersion"/"File)
		}
	  ListOfLastFiles | getline
	  }
	close(ListOfLastFiles)

	system("rm -f "LastZipFile)	# delete the last .zip
	print "...and now zipping that ->\t\t"LastVersion".zip"
	if (OS2)			# and make a new one by zipping all diffs
	    system("("Get_cd(DirLastVersion) " & zip -r -m -D -o "LastZipFile" *)")
	  else
	    system("("Get_cd(DirLastVersion) " ; zip -r -m -D -o "LastZipFile" *)")
	RemoveDir(DirLastVersion)	# remove the temporary dir of the last version
	} # end for real backup wrt the previous backup

  # now again for both the real and all backups: 
  # update or create the necessary management files

  # create ".LastVersion": it contains the CurrentVersion
  print CurrVersion"\n"TodayTime >LastVersionFile
  print "\n# please do NOT modify the contents of this file" >LastVersionFile
  print "# Deleting this file starts new branch (equiv. to the command backup_newbranch)" >LastVersionFile
  close(LastVersionFile)

  # sign this version and the time creation into the VersionsFile ".Versions"
  print CurrVersion"\t# "TodayTime >>VersionsFile
  close(VersionsFile)

  # Create file ",CurrVersion": it contains the successor of the CurrentVersion
  # note: the last version field (that after ".") must be number: else crazy result
  i=reverse_index_char(CurrVersion,".") # parse it wrt the last "."
  if (i==0) {	# the version name is a single number only
	S=CurrVersion; CurrVersion="" }
    else {	# the version name contains several number fields
	S=substr(CurrVersion,i+1); CurrVersion=substr(CurrVersion,1,i) }
  S++; if (length(S)==1) S="0"S
  CurrVersion=CurrVersion S # join the older trunk and the new subversion
  MakeCurrVersionFile(CurrVersion)
  print "\nNow you are working on the version:\t"CurrVersion
  print "Note: you can change it by editing the file\n      "CurrVersionFile
  print "Done."
  exit
  }
  # end of:  pmzbs backup _project_  and of  pmzbs backup_newbranch _project_ 


# pmzbs restore _project_ _version_
if (ARGV[1]=="restore") {
  if (ARGC<5) ERROR("'pmzbs restore' requires more arguments")
  # pmzbs restore _project_ _version_ _files_
  ProjectMustExist(Project,ProjectDir)
  Version=ARGV[3]
  if (!VersionExists(Version)) 
     ERROR(" version "Version" to be restored does not exist\n(see the file "VersionsFile" and "Version".zip).\n")
  
  # create the name of the output directory
  i=reverse_index_char(Project,"/")
  if (i==0) ToDir=Project"-"Version		 # no slash in the project name
    else ToDir=substr(Project,i+1)"-"Version # use only the part after the slash

  print "Restoring files from the project: "Project", version "Version
  print "The files are extracted to the directory "ToDir
  if (!DirectoryExist(ToDir)) MakeDir(ToDir)
  # unzip non-;diff files from the archive
  system(UNZIP" -qq -o "ProjectDir"/"Version".zip \""ARGV[4]"\" -x \"*;diff\" -d "ToDir" 2>"NULL)
  # now list the specified files
  ListFiles=UNZIP" -l "ProjectDir"/"Version".zip \""ARGV[4]";diff\""
  ListFiles | getline; ListFiles | getline; ListFiles | getline # skip 3 lines
  ListFiles | getline
  nAllVersions=0
  while (substr($1,1,3)!="---") {
    File=$ZIP_l_FILE  # File is "S1.txt" or "S1.txt;diff"
    if (index(File,";diff")!=0) { # the file is not in this archive
      TrueFileName=File; sub(";diff","",TrueFileName) # TrueFileName is "S1.txt"

      if (nAllVersion==0) { # read all versions superior or equal to the Version
	# now go to the successor of the Version
	do 
	  flag=getline1_nocomment(VersionsFile)
	while (flag>0 && $1!=Version)
	# now read all successors of the Version
	# flag=getline1_nocomment(VersionsFile)
	while (flag>0) {
	  S=ProjectDir"/"$1".zip"
	  if (!FileExist(S)) FATAL_ERROR("Version file "S" referred from "VersionsFile" does not exist!\n")
	  AllVersions[nAllVersions++]=S
	  flag=getline1_nocomment(VersionsFile)
	  }
	close(VersionsFile)
	if (nAllVersions==0) ERROR("no release has been found after "Version)
	##print "I have found RELEASES:"; for (i=0; i<nAllVersions; i++) print i": "AllVersions[i]
	} # end of reading AllVersions

      print "...restoring "TrueFileName
      # find  TrueFileName in a later version
      p=0;
      do {
	S=UNZIP" -l "AllVersions[p++]" "TrueFileName
	S | getline; S | getline; S | getline; S | getline; close(S)
	if (OS2) flag=(tolower($ZIP_l_FILE)!=tolower(TrueFileName))
	  else flag=($ZIP_l_FILE!=TrueFileName)
	}
      while (p<nAllVersions && flag)
      p--
      ## print "I have found "File" in p="p"  and "AllVersions[p]
      if (flag) BAD_ERROR("I have found the file "TrueFileName" in NO archive superior to "Version)

      # now unzip this true file name
      system(UNZIP" -o -q "AllVersions[p]" "TrueFileName" -d "ToDir)
      # and now patch it (if ;diff is not zero)
      rmNeeded=0
      for ( p--; p>=0; p--) {
	##print "PATCHING FROM p="p"  All="AllVersions[p]
	S=UNZIP" -l "AllVersions[p]" \""File"\""
	S | getline; S | getline; S | getline; S | getline; close(S)
	if ($1>0) { # the length of the ;diff is not zero and the patch can be applied
	  system(UNZIP" -o -q "AllVersions[p]" \""File"\" -d "ToDir)
	  system(GNUpatch" -n -s "ToDir"/"TrueFileName" <\""ToDir"/"File"\"")
	  rmNeeded=1
	  }
	}
      if (rmNeeded) # delete ;diff  and .orig
	system("rm -f \""ToDir"/"File"\" "ToDir"/"TrueFileName".orig")
      }
    ListFiles | getline
    }
  close(ListFiles)

  print "Done."
  exit
  }
  # end of: pmzbs backup _project_


ERROR("unknown pmzbs command: "ARGV[1])
}


# MakeVersion
# it returns VersionName by joining Version1, 2 and 3
# It sets these versions to 0 if there are empty string. It also 
##function MakeVersionName ( V1, V2, V3 ) {
##if (V1=="") V1=0; # well, this should not happen
##if (V2=="") { # name is formed only by Version1
##	if (length(V1)==1) V1="0"V1
##	return V1
##	}
##if (V3=="") { # name is formed by Version1 and Version2
##	if (length(V2)==1) V2="0"V2
##	return V1"."V2
##	}
### name is formed only by all Version1, Version2 and Version3 (all 3 levels)
##if (length(V3)==1) V3="0"V3
##return V1"."V2"."V3
##}


## crazy... this worked only five days...
## DirectoryExist
## return 1 if the specified directory exists
##function DirectoryExist ( Dir,   S ) {
##S="ls -a -1 -F "Dir"/. 2>"NULL" | grep \"/$\""
##S | getline
##close(S)
##return $0!=""
##}

# DirectoryExist
# return 1 if the specified directory exists
function DirectoryExist ( Dir,   S,flag ) {
S="ls -a -1 -F "Dir"/. 2>"NULL" | grep \"/$\""
flag=(S | getline); close(S)
return (flag>0)
}


# FileExist
# return 1 if the specified file exists
function FileExist ( File,   tmp ) {
tmp=(getline <File)
close(File)
# tmp=0 if eof (i.e. null file), +1 if sth has been read
return (tmp>=0)
}


# AreFilesTheSame
# return 1 if the specified files are the same (nothing returned by diff)
function AreFilesTheSame ( File1, File2,   Pipe,S ) {
Pipe="diff "File1" "File2
Pipe | getline S
close(Pipe)
return (S=="")
}


function MakeDir ( Dir ) {
if (BACKSLASHES) gsub("/","\\",Dir) # convert into the backslash format
system("mkdir "Dir)
}


function RemoveDir ( Dir ) {
if (BACKSLASHES) gsub("/","\\",Dir) # convert into the backslash format
system("rmdir "Dir)
}


# returns either "cd /tmp/X" or "cd \tmp\X" or "E: & cd E:\tmp\X"
function Get_cd ( Dir ) {
if (BACKSLASHES) {
  gsub("/","\\",Dir) # convert into the backslash format
  # separate drive name
  if (index(Dir,":")==2) 
    return substr(Dir,1,2)" & cd "Dir  # return e.g.  E: & cd E:\tmp\X
  }
return "cd "Dir
}


# reverse_index(s,t)  is similar to the build-in index(s,t), but it is reversed and t must be 1 char
function reverse_index_char ( s, t,   i ) {
for (i=length(s); i>0; i--)
  if (substr(s,i,1)==t) return i
return 0
}


# reverse_index(s,t)  is similar to the build-in index(s,t), but it is reversed
function reverse_index ( s, t ) {
Lt=length(t); Ls=length(s)
if (Lt==0) return Ls
if (Ls==0) return 0
print "FUNCTION reverse_index(s,t) HAS NOT BEEN WRITTEN\a!" >"/dev/stderr"
exit
}


# Error
function ERROR ( Hlaska ) {
print "\aERROR: "Hlaska"\n"; exit
}

# Bad Error
function BAD_ERROR ( Hlaska ) {
print "\aBAD ERROR: "Hlaska"\n";
}

# Fatal Error
function FATAL_ERROR ( Hlaska ) {
print "\aFATAL ERROR: "Hlaska"\n"; exit
}

function ProjectMustExist ( Project, ProjectDir ) {
if (!DirectoryExist(ProjectDir))
  ERROR("\tThe specified project "Project" (dir "ProjectDir")\n\thas not yet been created.")
}

# MakeCurrVersionFile creates new file CurrVersionFile, i.e. ";CurrVersion"
function MakeCurrVersionFile ( CurrVersion ) {
print CurrVersion >CurrVersionFile
print "\n# The first line of this file contains the name of the version to be backuped\n" >CurrVersionFile
print "# You can modify this file, but please be careful:" >CurrVersionFile
print "# the trunk of the version (everything up to the last point \".\" can be any word, but the " >CurrVersionFile
print "# last part (everything after the last point or the whole version name) must be a number." >CurrVersionFile
print "# You must avoid the version names already backuped (their list is in file "VersionsFile")" >CurrVersionFile
print "\n# Examples of correct version names:\n#   007  1.1.05  1.2.3.777  HelloDolly.01  auto.car.voiture.14" >CurrVersionFile
close(CurrVersionFile)
}

# VersionExists returns 1 if the specified Version already exists in VersionsFile, i.e. ",Versions"
function VersionExists ( Version,   i,flag ) {
do { # while something in the file
  flag=getline1_nocomment(VersionsFile)
  if ($1==Version) { # the Version has been found
	close(VersionsFile)
	return FileExist(ProjectDir"/"Version".zip")
	}
  }
while (flag>0)
# this Version has not been found
close(VersionsFile)
return 0
}

function ProcFor_list ( S,   LastVersion,LastVersionDate,LastVersionFile,tab ) {
LastVersionFile=Repository"/"S"/.LastVersion"
#print "Last="LastVersionFile
flag=(getline LastVersion < LastVersionFile)
flag=(getline LastVersionDate < LastVersionFile)
close(LastVersionFile)
if (length(S)<8) tab="\t\t"; else tab="\t"
if (flag>0) {
  S=S""tab"(last version "LastVersion
  if (length(LastVersionDate)>0) S=S"\ton "LastVersionDate
  S=S")"
  }
print S
}


# getline1_nocomment: reads $1, and skips lines with comments
function getline1_nocomment ( File,   flag,i ) {
do { # we must skip the lines with only the comment
  flag=(getline <File)
  if (flag<=0) break # eof
  if (NF==0) continue
  i=index($1,"#"); if (i==1) continue
  if (i>0) $1=substr($1,1,i-1) # delete the comment
  }
while (length($1)==0)
return flag
}


# getline_nocomment: reads $0, and skips lines with comments
function getline0_nocomment ( File,   flag,i ) {
do { # we must skip the lines with only the comment
  flag=(getline <File)
  if (flag<=0) break # eof
  if (NF==0) continue
  i=index($0,"#"); if (i==1) { $0=""; continue }
  if (i>0) $0=substr($0,1,i-1) # delete the comment
	# important note: NF and $1, $2,... are not affected!
  }
while (length($0)==0)
return flag
}


# Help

function Help () {
print "Syntax:"
print "pmzbs command [_project_] [options] [files]\n"
print "Commands:  new  list  backup  backup_newbranch  restore"
print "\npmzbs new _Project_  _version_levels_"
print "  Creates new Project in your repository."
print "  There can be 1, 2 or 3 version levels (e.g. 01, 0.01 or 0.0.01)"
print "\npmzbs list [_projects_]"
print "  lists all projects (or the projects specified by a regular expression"
print "  _projects_) in the repository"
print "\npmzbs backup _Project_"
print "  this backups your project into the version found in ,CurrVersion (make the"
print "    corresponding .zip file, puts its entry into the file ;Versions and"
print "    increases the current version number in the file ,CurrVersion)"
print "  The files to be backuped are listed in the file ,Files. The lines of this" 
print "    file are any valid commands of the INFO-zip program"
print "\npmzbs backup_newbranch _Project_"
print "  works like pmzbs backup, but only zips the files specified in ,Files and it"
print "  does not touch (doesn't diff) the previous version."
print "\npmzbs restore _Project_ _Version_ file(s)"
print "  unzips all file(s) from the version _Version_ of the project _Project_. If"
print "  there are diffs in that zip archive, then restore them by patching them from"
print "  the subsequent releases. Note: enclose wildcards in quotes, e.g. \"*\""
print "\nEnvironmental variables:"
print "  PMZBSROOT: absolute path to the repository (e.g. E:/home/petr/Save/pmzbs)"
print
}

# eof pmzbs.awk
